/** setlocale() for DJGPP with MS-DOS codepage support

    I don't claim this is bullet-proof ANSI setlocale,
      but some common cases like setlocale(LC_ALL, "")
      are handled well enough...

    Only LC_CTYPE category is supported now

    Author: Cyril Slobin <slobin@fe.msk.ru>

    Public Domain
    Made on Earth
*/

#include <dpmi.h>
#include <go32.h>
#include <inlines/ctype.ha>
#include <locale.h>
#include <string.h>

#define LC_SUPPORTED (LC_CTYPE)
#define UPCASE (__dj_ISALNUM | __dj_ISALPHA | __dj_ISGRAPH | __dj_ISPRINT | __dj_ISUPPER)
#define LOCASE (__dj_ISALNUM | __dj_ISALPHA | __dj_ISGRAPH | __dj_ISPRINT | __dj_ISLOWER)

char *setlocale(int category, const char *locale)
{
  static char *CLOCALE = "C";
  static char *CURRENT = "";
  static int state = 0;

  __dpmi_regs regs;
  struct { char id; unsigned short off, seg; } __attribute__ ((packed)) info;
  unsigned char buffer[0x82], lower, upper;
  int i;

  category &= LC_SUPPORTED;

  if (locale == NULL)
  {
    if ((category & state) == category)
      return CURRENT;
    if ((category & state) == 0)
      return CLOCALE;
    return NULL;
  }

  if (strcmp(locale, "") == 0)
  {
    if (category & LC_CTYPE)
    {
      regs.x.ax = 0x6502;
      regs.x.bx = 0xffff;
      regs.x.dx = 0xffff;
      regs.x.cx = 5;
      regs.x.es = __tb >> 4;
      regs.x.di = __tb & 0xf;

      __dpmi_int(0x21, &regs);

      if (regs.x.flags & 1)
        return NULL;

      dosmemget(__tb, 5, &info);
      dosmemget((info.seg << 4) + info.off, 0x82, buffer);

      if (*(short *)buffer != 0x80)
        return NULL;

      for (i = 0; i < 0x80; i++)
      {
        lower = i + 0x80;
        upper = (buffer+2)[i];
        if (lower != upper)
        {
          __dj_ctype_flags[lower+1] = LOCASE;
          __dj_ctype_toupper[lower+1] = upper;
          if (__dj_ctype_flags[upper+1] == 0)
            __dj_ctype_flags[upper+1] = UPCASE;
          if (__dj_ctype_tolower[upper+1] == upper)
            __dj_ctype_tolower[upper+1] = lower;
        }
      }
    }
    state |= category;
    return CURRENT;
  }

  if (strcmp(locale, "C") == 0 || strcmp(locale, "POSIX") == 0)
  {
    if (category & LC_CTYPE)
    {
      for (i = 0x80; i < 0x100; i++)
      {
        __dj_ctype_flags[i+1] = 0;
        __dj_ctype_toupper[i+1] = i;
        __dj_ctype_tolower[i+1] = i;
      }
    }
    state &= ~category;
    return CLOCALE;
  }

  return NULL;
}
